'''
Usage example:

    try:
        config: Config = Config.from_env_var()
        ...
    except ConfigError as e:
        ...
'''

import copy
import itertools
import os
import tomllib
from abc import ABC
from string import Template
from typing import Annotated, Iterator, Self

import pydantic as pd
from path import Path

from cstow.command import CmdVars

_CONFIG_PATH_ENV_VAR = 'CSTOW_CONFIG_PATH'

_CMD_TEMPLATE_DEFAULT = Template(
    "stow --${} --no-folding --verbose -t ${} -d ${} . ".format(*CmdVars.fields())
    + "2>&1 | grep -v -e '^BUG' -e '^WARN'"  # Clean output
)
_ROOT_DIR_DEFAULT = Path('/')


class ConfigError(ABC, Exception):
    '''Any error with a user config.'''


class ConfigEnvVarUnsetError(ConfigError):
    def __init__(self, env_var: str) -> None:
        super().__init__(
            f'Environment variable {env_var} is unset. Expected path to cstow_config.toml'
        )


class ConfigNotFoundError(ConfigError):
    def __init__(self, path: str) -> None:
        super().__init__(f'No such file: {path}\nExpected path to cstow_config.toml')


class InvalidConfigError(ConfigError):
    def __init__(self, path: str, message: str) -> None:
        super().__init__(f'Invalid config: {path}\n\n{message}')

    @classmethod
    def from_pydantic(cls, path: str, exception: pd.ValidationError) -> Self:
        '''Convert a Pydantic error message to an even more user-friendly one.'''
        messages: list[str] = []

        for err in exception.errors():
            loc = err['loc'][0]
            inp = err['input']
            msg = err['msg'].lstrip('Assertion failed, ')
            messages.append(f'{loc}\n    input: {inp}\n    error: {msg}')

        return cls(path, '\n\n'.join(messages))


def _validate_cmd_template(template_str: str) -> Template:
    '''Use with Pydantic and typing.Annotated.'''
    cmd_template = Template(template_str)

    assert cmd_template.is_valid(), "Invalid syntax in 'cmd_template'"
    assert set(cmd_template.get_identifiers()) == set(
        CmdVars.fields()
    ), f"'cmd_template' must contain all of these vars and no others: " + ", ".join(
        CmdVars.fields()
    )

    return cmd_template


def _expand(path: str) -> Path:
    return Path(path).expand()


CmdTemplate = Annotated[str, pd.AfterValidator(_validate_cmd_template)]
DirectoryPath = Annotated[pd.DirectoryPath, pd.BeforeValidator(_expand)]
TarsDirs = dict[DirectoryPath, list[pd.DirectoryPath]]


class Config(pd.BaseModel, extra='forbid'):
    '''
    The user config validated by Pydantic.

    Attributes:
        cmd_template: A template for constructing GNU Stow commands.
        root_dir: The root directory of source directories.
        targets_dirs: A dictionary of target directories and associated source directories.

    Raises:
        pydantic.ValidationError: If the config is invalid.
    '''

    cmd_template: CmdTemplate = _CMD_TEMPLATE_DEFAULT  # type: ignore
    root_dir: DirectoryPath = _ROOT_DIR_DEFAULT  # type: ignore
    targets_dirs: TarsDirs

    @pd.field_validator('targets_dirs', mode='before')
    @classmethod
    def _(
        cls, targets_dirs: dict[str, list[str]], info: pd.FieldValidationInfo
    ) -> dict[str, list[Path]]:
        '''Expand dirs in targets_dirs and prefix them with the root_dir.'''
        targets_dirs = copy.deepcopy(targets_dirs)

        assert 'root_dir' in info.data, "'root_dir' is invalid"
        root_dir = info.data['root_dir']

        assert isinstance(targets_dirs, dict), "'targets_dirs' must be a table"
        for target, dirs in targets_dirs.items():
            assert isinstance(dirs, list), "'dirs' must be an array"
            targets_dirs[target] = [root_dir / _expand(dir_) for dir_ in dirs]

        return targets_dirs  # type: ignore

    @classmethod
    def from_env_var(cls, env_var: str = _CONFIG_PATH_ENV_VAR) -> Self:
        '''
        Load a user config from a file provided by an environment variable.

        Args:
            env_var:
                An environment variable pointing to a user config file.
                Defaults to _CONFIG_PATH_ENV_VAR.

        Raises:
            ConfigEnvVarUnsetError: If the environment variable is unset.
            ConfigNotFoundError: If the file is not found.
            InvalidConfigError: If the config is invalid.
        '''
        return cls.from_path(_env_var_to_path(env_var))

    @classmethod
    def from_path(cls, path: Path) -> Self:
        '''
        Load a user config from a file.

        Args:
            path: A path to a user config file.

        Raises:
            ConfigNotFoundError: If the file is not found.
            InvalidConfigError: If the config is invalid.
        '''
        try:
            return cls(**tomllib.loads(path.read_text()))
        except FileNotFoundError:
            raise ConfigNotFoundError(path)
        except tomllib.TOMLDecodeError as e:
            raise InvalidConfigError(path, str(e))
        except pd.ValidationError as e:
            raise InvalidConfigError.from_pydantic(path, e)

    def each_target_and_dir(self) -> Iterator[tuple[DirectoryPath, DirectoryPath]]:
        '''
        Iterate over targets_dirs.

        Yields:
            Every pair of a target directory and a source directory.
        '''
        for target, dirs in self.targets_dirs.items():
            yield from itertools.product((target,), dirs)


def _env_var_to_path(env_var: str) -> Path:
    try:
        return Path(os.environ[env_var])
    except KeyError:
        raise ConfigEnvVarUnsetError(env_var)



import dataclasses as dc
import shlex
from enum import StrEnum, auto
from string import Template
from typing import Iterator

from path import Path


class CmdAction(StrEnum):
    '''GNU Stow actions.'''

    NO = auto()
    STOW = auto()
    RESTOW = auto()
    DELETE = auto()


class InvalidCmdActionError(Exception):
    def __init__(self, action: str) -> None:
        super().__init__(
            f"Action '{action}' is invalid\nUse one of these: " + ', '.join(CmdAction)
        )


def str_to_action(action: str) -> CmdAction:
    '''
    Validate a GNU Stow action.

    Raises:
        InvalidCmdActionError: If the GNU Stow action is invalid.
    '''
    if action in list(CmdAction):
        return CmdAction(action)
    raise InvalidCmdActionError(action)


@dc.dataclass(frozen=True)
class CmdVars:
    '''
    Variables for constructing GNU Stow commands from a template.

    Attributes:
        action: A GNU Stow action.
        target: A target directory.
        dir: A source directory.
    '''

    action: CmdAction
    target: Path
    dir: Path

    @classmethod
    def fields(cls) -> Iterator[str]:
        return (field.name for field in dc.fields(cls))

    def cmd(self, template: Template) -> str:
        '''Construct a GNU Stow command from a template.'''
        return template.substitute(
            **{field: shlex.quote(str(val)) for field, val in vars(self).items()}
        )



import sys
from typing import Any, NoReturn

import fire  # type: ignore

from cstow.command import CmdAction, InvalidCmdActionError, str_to_action
from cstow.config import Config, ConfigError
from cstow.stow import stow
from cstow.view import PlainView, RichView, View


def _error(*args: Any, **kwargs: Any) -> NoReturn:
    '''Print the error(s) and exit with 1.'''
    print('ERROR:\n', file=sys.stderr)
    print(*args, file=sys.stderr, **kwargs)
    exit(1)


def _cli(action: str = 'no', *, plain: bool = False) -> None:
    '''Use with Fire.'''
    view: View = PlainView() if plain else RichView()

    try:
        config: Config = Config.from_env_var()
        stow(str_to_action(action), config, view)
    except (ConfigError, InvalidCmdActionError) as e:
        _error(e)


def main() -> None:
    '''
    The entry point.

    Customize the help page, fire up the fire.Fire.
    '''

    _cli.__doc__ = f'''
    https://github.com/constkolesnyak/cstow/blob/main/README.md

    Args:
        action: Actions: {', '.join(CmdAction)}. 
        plain: Print plain text only. 
    '''

    fire.Fire(_cli)  # type: ignore


if __name__ == '__main__':
    _cli()



import subprocess
from typing import Callable

from path import Path

from cstow.command import CmdAction, CmdVars
from cstow.config import Config
from cstow.view import View

Proc = subprocess.CompletedProcess[bytes]
Run = Callable[[str], Proc]


def _run(cmd: str) -> Proc:
    '''
    Run a command in the shell.

    Raises:
        OSError: If the shell is not found, which is unlikely.
    '''
    return subprocess.run(args=[cmd], shell=True, capture_output=True)


def stow(action: CmdAction, config: Config, view: View, run: Run = _run) -> None:
    '''
    Commit an action on targets and dirs from a config, show the output using a view.

    Args:
        action: A GNU Stow action.
        config: A user config.
        view: A view.
        run: A Callable to run GNU Stow commands. Defaults to _run.
    '''
    view.show_action(action)

    for target, dir_ in config.each_target_and_dir():
        view.show_dir(Path(dir_))

        cmd: str = CmdVars(action, target, dir_).cmd(config.cmd_template)  # type: ignore
        proc: Proc = run(cmd)

        view.show_proc(proc)



import os
import sys
from abc import ABC, abstractmethod
from subprocess import CompletedProcess

from path import Path
from rich.console import Console
from rich.highlighter import RegexHighlighter
from rich.theme import Theme

from cstow.command import CmdAction


class View(ABC):
    '''The view for showing information to the user.'''

    @abstractmethod
    def show_action(self, action: CmdAction) -> None:
        raise NotImplementedError

    @abstractmethod
    def show_dir(self, dir_: Path) -> None:
        raise NotImplementedError

    @abstractmethod
    def show_proc(self, proc: CompletedProcess[bytes]) -> None:
        raise NotImplementedError


class PlainView(View):
    '''The view for printing plain text only.'''

    def show_action(self, action: CmdAction) -> None:
        print(f'@@@@@ {_action_to_title(action)} @@@@@\n')

    def show_dir(self, dir_: Path) -> None:
        print(f'### {dir_} ###\n')

    def show_proc(self, proc: CompletedProcess[bytes]) -> None:
        if stdout := proc.stdout.decode():
            print(stdout)
        if stderr := proc.stderr.decode():
            print(stderr, file=sys.stderr)


class _StowHighlighter(RegexHighlighter):
    """The highlighter for GNU Stow output."""

    highlights = [r'(?P<linking>(UN)?LINK)', r'(?P<linking>=>)']
    theme = Theme({"linking": "bold spring_green4"})


class RichView(View):
    '''The view for printing rich text.'''

    def __init__(self) -> None:
        self.console = Console(highlighter=_StowHighlighter(), theme=_StowHighlighter.theme)
        self.error_console = Console(stderr=True, style="red")

    def show_action(self, action: CmdAction) -> None:
        surround = '[deep_pink3]' + '@' * 15

        self.console.print(
            f'{surround}   [bold green]{_action_to_title(action)}   {surround}\n',
            justify='center',
        )

    def show_dir(self, dir_: Path) -> None:
        self.console.print(
            f"[cyan]{dir_.parent}{os.path.sep}[yellow bold]{dir_.name}\n", justify='center'
        )

    def show_proc(self, proc: CompletedProcess[bytes]) -> None:
        if stdout := proc.stdout.decode():
            self.console.print(stdout)
        if stderr := proc.stderr.decode():
            self.error_console.print('[bold]ERROR:\n')
            self.error_console.print(stderr)


def _action_to_title(action: CmdAction) -> str:
    return action.replace(CmdAction.NO, 'dry run').upper()
